"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const vue_eslint_parser_1 = require("vue-eslint-parser");
const embeddedSupport_1 = require("../../embeddedSupport/embeddedSupport");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const transformTemplate_1 = require("./transformTemplate");
const serviceHost_1 = require("./serviceHost");
function isVue(filename) {
    return path.extname(filename) === '.vue';
}
exports.isVue = isVue;
function parseVueScript(text) {
    const doc = vscode_languageserver_types_1.TextDocument.create('test://test/test.vue', 'vue', 0, text);
    const regions = embeddedSupport_1.getVueDocumentRegions(doc);
    const script = regions.getSingleTypeDocument('script');
    return script.getText() || 'export default {};';
}
exports.parseVueScript = parseVueScript;
function parseVueTemplate(text) {
    const doc = vscode_languageserver_types_1.TextDocument.create('test://test/test.vue', 'vue', 0, text);
    const regions = embeddedSupport_1.getVueDocumentRegions(doc);
    const template = regions.getSingleTypeDocument('template');
    if (template.languageId !== 'vue-html') {
        return '';
    }
    const rawText = template.getText();
    // skip checking on empty template
    if (rawText.replace(/\s/g, '') === '') {
        return '';
    }
    return rawText.replace(/ {10}/, '<template>') + '</template>';
}
function createUpdater(tsModule) {
    const clssf = tsModule.createLanguageServiceSourceFile;
    const ulssf = tsModule.updateLanguageServiceSourceFile;
    const scriptKindTracker = new WeakMap();
    const modificationTracker = new WeakSet();
    function isTSLike(scriptKind) {
        return scriptKind === tsModule.ScriptKind.TS || scriptKind === tsModule.ScriptKind.TSX;
    }
    function modifySourceFile(fileName, sourceFile, scriptSnapshot, version, scriptKind) {
        if (modificationTracker.has(sourceFile)) {
            return;
        }
        if (isVue(fileName) && !isTSLike(scriptKind)) {
            modifyVueScript(tsModule, sourceFile);
            modificationTracker.add(sourceFile);
            return;
        }
        if (serviceHost_1.isVirtualVueTemplateFile(fileName)) {
            // TODO: share the logic of transforming the code into AST
            // with the template mode
            const code = parseVueTemplate(scriptSnapshot.getText(0, scriptSnapshot.getLength()));
            const program = vue_eslint_parser_1.parse(code, { sourceType: 'module' });
            const tsCode = transformTemplate_1.getTemplateTransformFunctions(tsModule).transformTemplate(program, code);
            injectVueTemplate(tsModule, sourceFile, tsCode);
            modificationTracker.add(sourceFile);
        }
    }
    function createLanguageServiceSourceFile(fileName, scriptSnapshot, scriptTarget, version, setNodeParents, scriptKind) {
        const sourceFile = clssf(fileName, scriptSnapshot, scriptTarget, version, setNodeParents, scriptKind);
        scriptKindTracker.set(sourceFile, scriptKind);
        modifySourceFile(fileName, sourceFile, scriptSnapshot, version, scriptKind);
        return sourceFile;
    }
    function updateLanguageServiceSourceFile(sourceFile, scriptSnapshot, version, textChangeRange, aggressiveChecks) {
        const scriptKind = scriptKindTracker.get(sourceFile);
        sourceFile = ulssf(sourceFile, scriptSnapshot, version, textChangeRange, aggressiveChecks);
        modifySourceFile(sourceFile.fileName, sourceFile, scriptSnapshot, version, scriptKind);
        return sourceFile;
    }
    return {
        createLanguageServiceSourceFile,
        updateLanguageServiceSourceFile
    };
}
exports.createUpdater = createUpdater;
function modifyVueScript(tsModule, sourceFile) {
    const exportDefaultObject = sourceFile.statements.find(st => st.kind === tsModule.SyntaxKind.ExportAssignment &&
        st.expression.kind === tsModule.SyntaxKind.ObjectLiteralExpression);
    if (exportDefaultObject) {
        // 1. add `import Vue from 'vue'
        //    (the span of the inserted statement must be (0,0) to avoid overlapping existing statements)
        const setZeroPos = getWrapperRangeSetter(tsModule, { pos: 0, end: 0 });
        const vueImport = setZeroPos(tsModule.createImportDeclaration(undefined, undefined, setZeroPos(tsModule.createImportClause(tsModule.createIdentifier('__vueEditorBridge'), undefined)), setZeroPos(tsModule.createLiteral('vue-editor-bridge'))));
        const statements = sourceFile.statements;
        statements.unshift(vueImport);
        // 2. find the export default and wrap it in `__vueEditorBridge(...)` if it exists and is an object literal
        // (the span of the function construct call and *all* its members must be the same as the object literal it wraps)
        const objectLiteral = exportDefaultObject.expression;
        const setObjPos = getWrapperRangeSetter(tsModule, objectLiteral);
        const vue = tsModule.setTextRange(tsModule.createIdentifier('__vueEditorBridge'), {
            pos: objectLiteral.pos,
            end: objectLiteral.pos + 1
        });
        exportDefaultObject.expression = setObjPos(tsModule.createCall(vue, undefined, [objectLiteral]));
        setObjPos(exportDefaultObject.expression.arguments);
    }
}
/**
 * Wrap render function with component options in the script block
 * to validate its types
 */
function injectVueTemplate(tsModule, sourceFile, renderBlock) {
    // add import statement for corresponding Vue file
    // so that we acquire the component type from it.
    const setZeroPos = getWrapperRangeSetter(tsModule, { pos: 0, end: 0 });
    const vueFilePath = './' + path.basename(sourceFile.fileName.slice(0, -9));
    const componentImport = setZeroPos(tsModule.createImportDeclaration(undefined, undefined, setZeroPos(tsModule.createImportClause(setZeroPos(tsModule.createIdentifier('__Component')), undefined)), setZeroPos(tsModule.createLiteral(vueFilePath))));
    // import helper type to handle Vue's private methods
    const helperImport = setZeroPos(tsModule.createImportDeclaration(undefined, undefined, setZeroPos(tsModule.createImportClause(undefined, setZeroPos(tsModule.createNamedImports([
        setZeroPos(tsModule.createImportSpecifier(undefined, setZeroPos(tsModule.createIdentifier(transformTemplate_1.renderHelperName)))),
        setZeroPos(tsModule.createImportSpecifier(undefined, setZeroPos(tsModule.createIdentifier(transformTemplate_1.componentHelperName)))),
        setZeroPos(tsModule.createImportSpecifier(undefined, setZeroPos(tsModule.createIdentifier(transformTemplate_1.iterationHelperName)))),
        setZeroPos(tsModule.createImportSpecifier(undefined, setZeroPos(tsModule.createIdentifier(transformTemplate_1.listenerHelperName))))
    ])))), setZeroPos(tsModule.createLiteral('vue-editor-bridge'))));
    // wrap render code with a function decralation
    // with `this` type of component.
    const setRenderPos = getWrapperRangeSetter(tsModule, sourceFile);
    const statements = renderBlock.map(exp => tsModule.createStatement(exp));
    const renderElement = setRenderPos(tsModule.createStatement(setRenderPos(tsModule.createCall(setRenderPos(tsModule.createIdentifier(transformTemplate_1.renderHelperName)), undefined, [
        // Reference to the component
        setRenderPos(tsModule.createIdentifier('__Component')),
        // A function simulating the render function
        setRenderPos(tsModule.createFunctionExpression(undefined, undefined, undefined, undefined, [], undefined, setRenderPos(tsModule.createBlock(statements))))
    ]))));
    // replace the original statements with wrapped code.
    sourceFile.statements = setRenderPos(tsModule.createNodeArray([componentImport, helperImport, renderElement]));
    // Update external module indicator to the transformed template node,
    // otherwise symbols in this template (e.g. __Component) will be put
    // into global namespace and it causes duplicated identifier error.
    sourceFile.externalModuleIndicator = componentImport;
}
/** Create a function that calls setTextRange on synthetic wrapper nodes that need a valid range */
function getWrapperRangeSetter(tsModule, wrapped) {
    return (wrapperNode) => tsModule.setTextRange(wrapperNode, wrapped);
}
//# sourceMappingURL=preprocess.js.map